import re

# Each regex matches with 2 groups: 1) text to split  2) what to split on
# The regex for [Gent,groen] will result in:
# group1 = 'Gent,groen' and group2= ','
# The regex for Gent & groen will result in:
# group1 = 'Gent & groen' and group2= '&'
from deprecated import deprecated

AND_FILTER_REGEXES = [re.compile("([^&]+(&)(?:[^&]+&?)*)")]  # Matches Gent & groen
OR_FILTER_REGEXES = [re.compile(r"\[((?:[^,]+(,)?)+)\]")]  # Matches [Gent,groen]


class QueryBuilder(object):
    def __init__(self, query_params=None, es_version="1"):
        """
        Create a query builder.

        :param query_params: dict with query parameters
        """
        self.es_version = es_version
        self.query_params = query_params
        self.query = self._build_initial_query_string()
        self.filters = []

    def _build_initial_query_string(self):
        if (
            self.query_params is None
            or "query" not in self.query_params
            or self.query_params["query"] == ""
        ):
            return {"match_all": {}}
        else:
            all_fields = "_all" if self.es_version == "1" else "*"
            return {
                "query_string": {
                    "default_field": all_fields,
                    "query": self.query_params["query"],
                }
            }

    @deprecated(version="1.5.0", reason="Integrated in add_query_param_filters()")
    def add_named_concept_filters(self, named_filter_concepts):
        """
        Add named concept filters.

        :param named_filter_concepts: dict with named filter concepts which will be
        mapped as the key as query param and the value as search string
        """
        for concept_key, concept_name in named_filter_concepts.items():
            self.add_concept_filter(concept_key, concept_name=concept_name)

    def add_query_param_filters(self, query_param_names, query_param_translations=None):
        """
        Add query param filters.

        :param query_param_names: list with filter concepts
        :param query_param_translations: translates query_param to es_field
        """
        for query_param in query_param_names:
            if query_param_translations:
                es_field = query_param_translations.get(query_param)
            else:
                es_field = None
            self.add_query_param_filter(query_param, es_field=es_field)

    @deprecated(
        version="1.5.0", reason="Use function add_query_param_filters() instead"
    )
    def add_concept_filters(self, filter_concepts):
        """
        Add concept filters.

        :param filter_concepts: list with filter concepts
        """
        for concept in filter_concepts:
            self.add_concept_filter(concept)

    def add_query_param_filter(self, query_param_name, es_field=None):
        """
        Add a query_param filter.

        :param query_param_name: concept which will be used as
        lowercase string in a search term
        :param es_field: name of the place where there will be searched for
        """
        if query_param_name in self.query_params.keys():
            if not es_field:
                es_field = query_param_name
            if isinstance(self.query_params[query_param_name], list):
                es_filter = {"bool": {"should": []}}
                for or_filter in self.query_params[query_param_name]:
                    es_filter["bool"]["should"].append(
                        self._build_concept_term(es_field, or_filter)
                    )
            else:
                es_filter = self._build_concept_term(
                    es_field, self.query_params[query_param_name]
                )
            self.filters.append(es_filter)

    @deprecated(version="1.5.0", reason="Use function add_query_param_filter() instead")
    def add_concept_filter(self, concept, concept_name=None):
        """
        Add a concept filter.

        :param concept: concept which will be used as lowercase string in a search term
        :param concept_name: name of the place where there will be searched for
        """
        if concept in self.query_params.keys():
            if not concept_name:
                concept_name = concept
            if isinstance(self.query_params[concept], list):
                if self.es_version == "1":
                    es_filter = {"or": []}
                    for or_filter in self.query_params[concept]:
                        es_filter["or"].append(
                            self._build_concept_term(concept_name, or_filter)
                        )
                else:
                    es_filter = {"bool": {"should": []}}
                    for or_filter in self.query_params[concept]:
                        es_filter["bool"]["should"].append(
                            self._build_concept_term(concept_name, or_filter)
                        )
            else:
                es_filter = self._build_concept_term(
                    concept_name, self.query_params[concept]
                )
            self.filters.append(es_filter)

    def _build_query_param_term_filter(self, es_field, query_param_value):
        return {"term": {es_field: str(query_param_value).lower()}}

    def _build_concept_term(self, concept_name, concept):
        filter_method = "term" if str(self.es_version) == "1" else "match"
        return {filter_method: {concept_name: str(concept).lower()}}

    def parse_full_text_search_param(self, value):
        """
        Convert the user queryparam input to a valid query_string for ES.

        This does not really support quotes yet.

        Onroerend erfgoed has some custom rules as to how users can enter a
        query. For example, the user can use `&` to perform an "AND" query.
        This method will turn all the special cases from onroerend erfgoed
        into a valid query for Elasticsearch by, for example, turning "x & y"
        into "x + y".
        """
        result = self.parse_full_text_search_param_iterable(value)
        if result is not None:
            return result
        result = self.parse_full_text_search_param_or_patterns(value)
        if result is not None:
            return result
        result = self.parse_full_text_search_param_and_patterns(value)
        if result is not None:
            return result
        return value.strip()

    def parse_full_text_search_param_iterable(self, value):
        """
        Translate a list or tuple into AND for ES.

        Sending ?text=Gent&text=groen has to be seen as an AND.
        This method will transform the value ["Gent", "Groen"]
        into "(Gent) + (groen)".
        """
        if isinstance(value, list) or isinstance(value, tuple):
            filters = [self.parse_full_text_search_param(val) for val in value]
            return " + ".join("(" + parsed + ")" for parsed in filters)
        return None

    def parse_full_text_search_param_or_patterns(self, value):
        """
        Look for specific OR patterns and translate them to ES.

        Sending text=[Gent,groen] has to be seen as an OR.
        This method will transform the string "[Gent,groen]"
        into "Gent | groen"
        """
        for regex in OR_FILTER_REGEXES:
            match = regex.match(value)
            if match:
                values = self.quote_ignore_split(match)
                filters = [self.parse_full_text_search_param(val)
                           if not (value.startswith('"') and value.endswith('"'))
                           else val
                           for val in values]
                return " | ".join(filters)
        return None

    def parse_full_text_search_param_and_patterns(self, value):
        """
        Look for specific AND patterns and translate them to ES.

        Sending text=Gent & groen has to be seen as an AND.
        This method will transform the string "Gent & Groen"
        into "Gent + groen"
        """
        for regex in AND_FILTER_REGEXES:
            match = regex.match(value)
            if match:
                values = self.quote_ignore_split(match)
                filters = [self.parse_full_text_search_param(val)
                           if not (value.startswith('"') and value.endswith('"'))
                           else val
                           for val in values]
                return " + ".join(filters)
        return None

    def quote_ignore_split(self, match):
        to_split = match.group(1)
        splitter = match.group(2)
        lijst = [element for element in re.split('("[^"]*")', to_split) if element != '']
        final_list = []
        for element in lijst:
            if element.startswith('"') and element.endswith('"'):
                final_list.append(element)
            else:
                final_list.extend(
                    [element.strip() for element in element.strip().split(splitter)
                     if element != ''])
        return final_list

    def set_full_text_search_query(
        self,
        field_boosts,
        query,
        default_operator="AND",
        flags="AND|OR|NOT|PREFIX|PHRASE|PRECEDENCE|ESCAPE|WHITESPACE",
        analyze_wildcard=True,
    ):
        """
        Build a simple_query_string query.

        The fields will be taken from the `field_boosts` keys. And each field
        will receive a score boost from `field_boosts` values.
        This is not a filter but sets the query itself because we care about
        scores.

        :type field_boosts: dict[str, int]
        :param field_boosts: fields and their score boost values, 1 default.
        :type query: str
        :param query: the simple_query_string query.
        :type default_operator: str
        :param default_operator: default_operator of the simple_query_string.
           Set this to 'AND' or 'OR'
        :type flags: str
        :param flags: flags of the simple_query_string. By default this is
           all the flags except fuzzyness.
        :type analyze_wildcard: bool
        :param analyze_wildcard: analyze_wildcard of the simple_query_string.
           Setting this to true will allow prefix wildcard strings to work.
        """
        # analyze_wildcard true will make it so when a prefix wildcard is used,
        # the prefix is still analyzed. "pre-fix*" without analyze would make
        # ES look for the term "pre-fix". But by default the word would
        # (should) have been broken up into terms 'pre' and 'fix'. With
        # analyze true. ES will look for "pre" and "fix*".
        # ^x are the score boost numbers.
        fields = ["{}^{}".format(field, boost) for field, boost in field_boosts.items()]
        self.query = {
            "simple_query_string": {
                "fields": fields,
                "query": query,
                "default_operator": default_operator,
                "flags": flags,
                "analyze_wildcard": analyze_wildcard,
            }
        }

    def build(self):
        """
        Build the query string, which can be used for a search query.

        :return: the query string
        """
        if self.es_version == "1":
            if len(self.filters) > 0:
                return {
                    "filtered": {"query": self.query, "filter": {"and": self.filters}}
                }
            else:
                return self.query
        else:
            query = {"bool": {"must": self.query}}
            if len(self.filters) > 0:
                query["bool"]["filter"] = self.filters
            return query

    def and_(self, *filters):
        return {"bool": {"must": filters}}

    def or_(self, *filters):
        return {"bool": {"should": filters}}

    def not_(self, *filters):
        return {"bool": {"must_not": filters}}

    def field_exists(self, field_name):
        return {"exists": {"field": field_name}}



# -*- coding: utf-8 -*-

from oe_utils.data.data_transfer_objects import ResultDTO
from oe_utils.data.models import Wijziging


class DataManager(object):
    """A datamanager base class."""

    @staticmethod
    def process_ranged_query(query, result_range):
        """
        Create a query with limit and offset.

        :param query: the query to be processed
        :param result_range: :class: 'oe_utils.range_parser.Range'
        :return: :class:`oe_utils.data.data_transfer_objects.ResultDTO`
        """
        total = query.count()
        if result_range is not None:
            data = (
                query.offset(result_range.start)
                .limit(result_range.get_page_size())
                .all()
            )
        else:
            data = query.all()
        return ResultDTO(data, total)

    @staticmethod
    def ordered_result_list(query, model_type, sortlist):
        for sorttuple in sortlist:
            if hasattr(model_type, sorttuple[0]):
                attr = getattr(model_type, sorttuple[0])
                query = (
                    query.order_by(attr.asc())
                    if sorttuple[1] == "asc"
                    else query.order_by(attr.desc())
                )
        return query

    def __init__(self, session, cls):
        """
        Create a datamanager for a specific model class.

        :param session: a db session
        :param cls: the class of the objects to manage
        :return:
        """
        if not session.autoflush:
            raise ValueError("Session autoflush must be True.")
        self.session = session
        self.cls = cls

    def get_one(self, object_id):
        """
        Retrieve an object by its object_id, raises when not found.

        :param object_id: the objects id.
        :return: the requested object
        :raises: :class: NoResultFound when the object could not be found
        """
        return self.session.query(self.cls).filter_by(id=object_id).one()

    def get_one_for_update(self, object_id):
        """
        Retrieve an object by its object_id and locks the row.

        Does a select for update on the row,
        results in row level lock for the duration of the transaction

        :param object_id: the objects id.
        :return: the requested object
        :raises: :class: NoResultFound when the object could not be found
        """
        return (
            self.session.query(self.cls).with_for_update().filter_by(id=object_id).one()
        )

    def get(self, object_id, cls=None):
        """
        Retrieve an object by its object_id or None when not found.

        :param: object_id: the objects id.
        :param: cls: the objects class,
        if None use the default class from the datamanager
        :return: the requested object or None if not found
        """
        cls = self.cls if cls is None else cls
        return self.session.query(cls).get(object_id)

    def get_for_update(self, object_id, cls=None):
        """
        Retrieve an object by its object_id and locks the row.

        Does a select for update on the row,
        results in row level lock for the duration of the transaction

        :param: object_id: the objects id.
        :param: cls: the objects class,
        if None use the default class from the datamanager
        :return: the requested object or None if not found
        """
        cls = self.cls if cls is None else cls
        return self.session.query(cls).with_for_update().get(object_id)

    def delete(self, object_id):
        """
        Delete an object by its id.

        :param object_id: the objects id.
        :return: the deleted object
        :raises: :class: NoResultFound when the object could not be found
        """
        obj = self.session.query(self.cls).filter_by(id=object_id).one()
        self.session.delete(obj)
        return obj

    def save(self, obj):
        """
        Save an object.

        :param obj: the object
        :return: the saved object
        """
        if obj not in self.session:
            self.session.add(obj)
        else:
            obj = self.session.merge(obj)
        self.session.flush()
        self.session.refresh(obj)
        return obj


class FeedArchiveNotFound(Exception):
    pass


class AtomFeedManager(object):
    """A data manager for feeds."""

    def __init__(self, session, feed_entries_per_page, feedentry_model):
        """
        Create an atom feed manager.

        :param session: a db session
        :param feed_repository: a repository to store the archived feeds
        """
        self.session = session
        self.feedentry_model = feedentry_model
        self._current_feed = None
        self.feed_entries_per_page = feed_entries_per_page

    def _get_max_min(self):
        return self.session.execute(
            """
            SELECT max(feed_id), min(feed_id) FROM (
            SELECT (row_number () OVER (ORDER BY id) -1) / {per_page} AS "feed_id"
            FROM feedentries
            ) AS "feed_ids" """.format(per_page=self.feed_entries_per_page)
        ).fetchall()[0]

    @property
    def current_feed_id(self):
        max, min = self._get_max_min()
        if self._current_feed is None:
            self._current_feed = max
        return self._current_feed

    def get_sibling_id(self, feed_id, sibling_type):
        """
        Get a previous/next sibling from a feed.

        :param feed_id: id of the feed
        :param sibling_type: sibling type ('previous', 'next')
        :return: the sibling
        """
        max, min = self._get_max_min()
        if sibling_type == "previous":
            if feed_id - 1 >= min:
                return feed_id - 1
            else:
                return None
        elif sibling_type == "next":
            if feed_id + 1 <= max:
                return feed_id + 1
            else:
                return None
        else:
            raise Exception("Unhandled sibling relation type")  # pragma no cover

    def save_object(self, obj):
        """
        Save an object to the db.

        :param obj: the object to save
        :return: the saved object
        """
        self.session.add(obj)
        self.session.flush()
        return obj

    def get_atom_feed_entry(self, feedentry_id):
        """
        Get a specific feed entry.

        :param feedentry_id: id of the feed entry to retrieve
        :return: the feed entry
        """
        return (
            self.session.query(self.feedentry_model)
            .filter(self.feedentry_model.id == feedentry_id)
            .one()
        )

    def get_feed_entries_where_feed_id(self, feed_id):
        """
        Get all entries linked to a feed_id.

        :return: feed entries of one feed_id (page)
        """
        return self.session.execute(
            """
            SELECT * FROM (
            SELECT *, (row_number () over (order by id) -1) / {per_page} as "feed_id"
            FROM feedentries order by id
            ) f
            WHERE feed_id = {feed_id} """.format(
                per_page=self.feed_entries_per_page, feed_id=feed_id
            )
        ).fetchall()

    def get_all_entries(self):
        """
        Get all entries with generated feed_id column.

        :return: all entries with feed_id column
        """
        return self.session.execute(
            """
            SELECT *, (row_number () over (order by id) -1) / {per_page} as "feed_id"
            FROM feedentries
            ORDER by id """.format(per_page=self.feed_entries_per_page)
        ).fetchall()


class AuditManager(DataManager):
    """A data manager for audit revisions."""

    def __init__(self, session):
        """
        Create an audit manager.

        :param session: a db session
        """
        super(AuditManager, self).__init__(session, Wijziging)
        self.session = session

    @staticmethod
    def create_revision():
        """
        Create a new revision object.

        :return: new revision object
        """
        return Wijziging()



# -*- coding: utf-8 -*-
import logging

from cornice import validators

from oe_utils.validation import ValidationFailure

LOG = logging.getLogger(__name__)


def schema_validator(request, schema=None, deserializer=None, **kwargs):
    """
    Validate the request against the configured schema.

    This method exists because the default cornice validators do not bind the
    request to the schema, and our schemas need it often. For example to get
    the database session to validate database ids.
    """
    if schema is None:
        return
    schema = schema.bind(request=request)
    validators.colander_validator(request, schema=schema,
                                  deserializer=deserializer, **kwargs)


def cornice_error_handler(request):
    """
    Handle validation errors in a OE-compatible way.

    This mimics the behaviour in other applications who use manual validation
    The 2nd parameter of ValidationFailure in other applications is normally
    `colander.Invalid.asdict()`. Here we replicate this output from the errors
    given by cornice in `request.errors`.
    """
    service_name = request.current_service.name
    raise ValidationFailure(
        "Fouten bij het valideren van de {} request.".format(service_name),
        {
            error["location"] + "." + error["name"]: error["description"]
            for error in request.errors
        },
    )



# -*- coding: utf-8 -*-
from deprecated import deprecated


@deprecated(
    version="1.5.0",
    reason="This method is duplicated at oe_utils.search.parse_sort_string.",
)
def parse_sort_string(sort):
    """
    Parse a sort string for use with the db.

    :param: sort: the sort string
    """
    if sort is None:
        return []
    else:
        sort_values = sort.rsplit(",")
        sortlist = []
        for se in sort_values:
            order = "desc" if se[0:1] == "-" else "asc"
            field = se[1:] if se[0:1] in ["-", "+"] else se
            field = field.strip()
            sortlist.append((field, order))
        return sortlist



# -*- coding: utf-8 -*-


class ResultDTO:
    """A DTO for returning results."""

    def __init__(self, data, total, aggregations=None):
        self.data = data
        self.total = total
        self.aggregations = aggregations



"""
Dit bestand bevat een OE Base Model dat in alle OE toepassingen kan worden gebruikt.

Het voordeel van een generieke Base is dat we op deze werkwijze per toepassing 1 zelfde
Base kunnen gebruiken voor al de aanwezige Models.
Dit omvat zowel Models van de applicatie zelf,
als de Models uit libraries zoals oe_utils en oe_geoutils.

Hier kan eveneens de nodige generieke metadata aan ons Base Model worden meegegeven.
Hiervoor kan sqlalchemy.schema.MetaData worden gebruikt
"""
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.sql import func

Base = declarative_base()


class Wijziging(Base):
    """
    A database table configuration object.

    This table contains information about the audit of a resource object.

    This object will not create the db table object.
    To create the table insert following code in the alembic migration file

    `alembic revision -m "wijzigingshistoriek"`


    from alembic import op
    import sqlalchemy as sa
    from sqlalchemy.dialects.postgresql import JSON
    from sqlalchemy.sql import func


    def upgrade():
        op.create_table(
            "wijzigingshistoriek",
            sa.Column("versie", sa.String(), nullable=False),
            sa.Column("resource_object_id", sa.Integer(), nullable=False),
            sa.Column(
                "updated_at",
                sa.DateTime(timezone=True),
                default=func.now(),
                nullable=False,
            ),
            sa.Column("updated_by", sa.String(length=255), nullable=False),
            sa.Column("updated_by_omschrijving", sa.String(length=255), nullable=False),
            sa.Column("resource_object_json", JSON, nullable=True),
            sa.Column("actie", sa.String(length=50), nullable=False),
            sa.PrimaryKeyConstraint("versie", name="wijzigingshistoriek_pk"),
        )

        op.execute("GRANT ALL ON wijzigingshistoriek to <user>_dml")


    def downgrade():
        op.drop_table("wijzigingshistoriek")
    """

    __tablename__ = "wijzigingshistoriek"
    versie = Column(String(64), primary_key=True)
    resource_object_id = Column(Integer, nullable=False)
    updated_at = Column(DateTime(timezone=True), default=func.now(), nullable=False)
    updated_by = Column(
        String(255), default="https://id.erfgoed.net/actoren/501", nullable=False
    )
    updated_by_omschrijving = Column(
        String(255), default="Onroerend Erfgoed", nullable=False
    )
    resource_object_json = Column(MutableDict.as_mutable(JSON()), nullable=True)
    actie = Column(String(50), nullable=False)



# -*- coding: utf-8 -*-
import logging

from sqlalchemy import event
from sqlalchemy.orm import object_session
from zope.interface import Interface

from oe_utils.jobs import queue_job

log = logging.getLogger()


class IIndexer(Interface):
    pass


class Indexer(object):
    """
    Handle the indexing of the DB to ES.

    This object contains a number listeners.
    First a list of the changes to be made will be kept.
    In case of a commit, these changes will also be executed.
    """

    def __init__(self, settings, index_operation, index_operation_name, cls):
        self.sessions = set()
        self.index_operation = index_operation
        self.index_operation_name = index_operation_name
        self._register_event_listeners(cls)
        self.settings = settings
        self.cls_name = cls.__name__

    def _register_event_listeners(self, cls):
        """
        Register event listeners.

        :param cls: DB class
        """
        event.listen(cls, "after_insert", self._new_listener)
        event.listen(cls, "after_update", self._update_listener)
        event.listen(cls, "after_delete", self._delete_listener)

    @staticmethod
    def _update_listener(mapper, connection, target):
        _add_to_session_list(target, operation="UPDATE")

    @staticmethod
    def _new_listener(mapper, connection, target):
        _add_to_session_list(target, operation="ADD")

    @staticmethod
    def _delete_listener(mapper, connection, target):
        _add_to_session_list(target, operation="REMOVE")

    def register_session(self, session, redis=None):
        session.redis = redis
        session.index_new = session.index_new if hasattr(session, "index_new") else {}
        session.index_new[self.cls_name] = set()
        session.index_dirty = (
            session.index_dirty if hasattr(session, "index_dirty") else {}
        )
        session.index_dirty[self.cls_name] = set()
        session.index_deleted = (
            session.index_deleted if hasattr(session, "index_deleted") else {}
        )
        session.index_deleted[self.cls_name] = set()
        self.sessions.add(session)
        event.listen(session, "after_commit", self.after_commit_listener)
        event.listen(session, "after_rollback", self.after_rollback_listener)

    def after_commit_listener(self, session):
        """
        Process the changes.

        All new or changed items are now indexed.
        All deleted items are now removed from the index.
        """
        log.info("Commiting indexing orders for session %s" % session)
        try:
            if not any(
                (
                    session.index_new[self.cls_name],
                    session.index_dirty[self.cls_name],
                    session.index_deleted[self.cls_name],
                )
            ):
                return
            if session.redis is not None:
                queue_job(
                    session.redis,
                    self.settings["redis.queue_name"],
                    self.index_operation_name,
                    session.index_new[self.cls_name],
                    session.index_dirty[self.cls_name],
                    session.index_deleted[self.cls_name],
                    self.settings,
                )
            else:
                log.info(
                    "Redis not found, "
                    "falling back to indexing synchronously without redis"
                )
                self.index_operation(
                    session.index_new[self.cls_name],
                    session.index_dirty[self.cls_name],
                    session.index_deleted[self.cls_name],
                    self.settings,
                )
            session.index_new[self.cls_name].clear()
            session.index_dirty[self.cls_name].clear()
            session.index_deleted[self.cls_name].clear()
        except AttributeError:
            log.warning(
                "Trying to commit indexing orders, but indexing sets are not present."
            )

    def after_rollback_listener(self, session):
        """
        Rollback of the transaction, undo the indexes.

        If our transaction is terminated, we will reset the
        indexing assignments.
        """
        log.info("Removing indexing orders.")
        try:
            session.index_new[self.cls_name].clear()
            session.index_dirty[self.cls_name].clear()
            session.index_deleted[self.cls_name].clear()
        except (AttributeError, KeyError):
            log.warning(
                "Trying to remove indexing orders, but indexing sets are not present."
            )

    def remove_session(self, session):
        """
        Remove a session from the indexer.

        :param sqlalchemy.session.Session session: Database session to remove
        """
        try:
            del session.redis
        except AttributeError:
            pass
        try:
            del session.index_new[self.cls_name]
            del session.index_dirty[self.cls_name]
            del session.index_deleted[self.cls_name]
        except (AttributeError, KeyError):
            log.warning("Removing a session that has no indexing sets.")
        self.sessions.remove(session)


def _add_to_session_list(target, operation):
    session = object_session(target)
    try:
        if operation == "ADD":
            session.index_new[target.__class__.__name__].add(target.id)
        elif operation == "UPDATE":
            session.index_dirty[target.__class__.__name__].add(target.id)
        elif operation == "REMOVE":
            session.index_deleted[target.__class__.__name__].add(target.id)
        log.info(operation + ": " + str(target) + " {0} from index".format(target.id))
    except (AttributeError, KeyError):
        log.warning(
            "Trying to register a "
            + str(target)
            + " for indexing "
            + operation
            + ", but indexing sets are not present."
        )



# -*- coding: utf-8 -*-
import logging

from rq import Connection
from rq import Queue


def queue_job(redis_connection, queue_name, delegate, *args):
    """
    Queue a job.

    :param redis_connection: redis connection instance
    :param queue_name: name of the queue to place the job on
    :param delegate: function to be executed by the queue -> 'package.module.function'
    :param args: arguments list used by called function
    :return: id of the job
    """
    logging.info("scheduling job {}".format(delegate))
    if callable(delegate):
        delegate = delegate.__module__ + "." + delegate.__name__
    with Connection(redis_connection):
        q = Queue(queue_name)
        job = q.enqueue(delegate, *args)
        return job.id



import json
from io import BytesIO

from reportlab.graphics import renderPDF
from reportlab.graphics import shapes
from reportlab.graphics.barcode.qr import QrCodeWidget
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus.paragraph import Paragraph


class QrPdf(object):
    def __init__(
        self,
        request,
        id,
        onderwerp,
        base_url,
        route,
        uri,
        form_data,
        pagesize=A4,
        qr_size=5 * cm,
    ):
        super(QrPdf, self).__init__()
        self.request = request
        self.pagesize = pagesize
        self.qr_size = qr_size
        self.id = id
        self.onderwerp = onderwerp
        self.form_data = form_data
        self.route = route
        self.base_url = base_url
        self.uri = uri

    def generate_pdf(self):
        in_memory_file = BytesIO()
        canvas = Canvas(in_memory_file, pagesize=self.pagesize)
        self._add_text(canvas)
        self._add_qr_code(canvas)
        canvas.save()
        in_memory_file.seek(0)
        return in_memory_file

    def _add_text(self, canvas):
        width, height = self.pagesize
        text = ("Onderwerp: {subject}<br/>URI: {uri}"
                .format(subject=self.onderwerp, uri=self.uri))
        style = ParagraphStyle('Normal')
        style.fontName = 'Courier'
        paragraph = Paragraph(text, style)
        paragraph.wrap(width - (4 * cm), 10 * cm)  # required to wrap before draw
        paragraph.drawOn(canvas, 2 * cm, height - 2 * cm)  # height 0 is the bottom

    def _add_qr_code(self, canvas):
        width, height = self.pagesize
        url = self.base_url + self.request.route_path(self.route, id=self.id)
        qr_data = json.dumps({"url": url, "form_data": self.form_data})

        drawing = shapes.Drawing()
        qr = QrCodeWidget(value=qr_data, barHeight=self.qr_size, barWidth=self.qr_size)
        drawing.add(qr)
        renderPDF.draw(
            drawing, canvas, width / 2 - self.qr_size / 2, height / 2 - self.qr_size / 2
        )



# -*- coding: utf-8 -*-
import colander
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.httpexceptions import HTTPNotFound

from oe_utils.validation import ValidationFailure


def get_resource(data_manager, resource_id, detail="No details provided."):
    resource = data_manager.get(resource_id)
    if not resource:
        raise HTTPNotFound(detail=detail)
    return resource


def get_json_from_request(request):
    try:
        return request.json_body
    except AttributeError as e:
        raise HTTPBadRequest(detail="Request bevat geen json body. \n%s" % e)
    except ValueError as e:  # pragma no cover
        raise HTTPBadRequest(
            detail="Request bevat incorrecte json body. \n%s" % e
        )  # pragma no cover


def validate(request, json_data, mapping_schema):
    validaton_schema = mapping_schema.bind(request=request)
    try:
        return validaton_schema.deserialize(json_data)
    except colander.Invalid as e:
        raise ValidationFailure("Fouten bij het valideren van request.", e.asdict())


class RouteConfigArgs:
    def __init__(
        self,
        response_schemas,
        schema=None,
        permission=None,
        tags=None,
        protected=False,
        **kwargs
    ):
        self.tags = tags
        self.response_schemas = response_schemas
        self.schema = schema
        self.permission = permission
        if protected:
            kwargs["protected"] = protected
        self.__dict__.update(kwargs)


class RouteConfig:
    def __init__(
        self, service, get_args=None, post_args=None, put_args=None, delete_args=None
    ):
        self.service = service
        self.get_args = get_args
        self.post_args = post_args
        self.put_args = put_args
        self.delete_args = delete_args

    @property
    def post(self):
        return self.service.post(**self.post_args.__dict__)

    @property
    def get(self):
        return self.service.get(**self.get_args.__dict__)

    @property
    def put(self):
        return self.service.put(**self.put_args.__dict__)

    @property
    def delete(self):
        return self.service.delete(**self.delete_args.__dict__)

    @property
    def get_validation_scheme(self):
        return self.get_args.schema.__class_schema_nodes__[-1]

    @property
    def put_validation_scheme(self):
        return self.put_args.schema.__class_schema_nodes__[-1]

    @property
    def post_validation_scheme(self):
        return self.post_args.schema.__class_schema_nodes__[-1]



import colander
import requests
import rfc3987


class ValidationFailure(Exception):
    def __init__(self, msg, errors):
        self.msg = msg
        self.errors = errors


def sqlalchemy_validator(node, value, model, session=None, msg="%s is ongeldig"):
    """
    Validate a sqlalchemy object or identifier.

    Function checks if value is an instance of sqlalchemy model
    or an id of one.
    :param node: A colander SchemaNode.
    :param value: The value to be validated. Is the key of the
        SQLAlchemy model.
    :param model: A SQLAlchemy model.
    :param session: A SQLAlchemy session. Is required if value is not an
        instance of a SQLAlchemy model.
    :param msg: A msg to attach to a `colander.Invalid` exception.
    :raises colander.Invalid: If not a valid sqlalchemy model.
    """
    m = session.query(model).get(value)

    if not m:
        raise colander.Invalid(node, msg % value)

    return m


#########################################################################################
# BASIC DATA TYPES (ENRICHED FOR O.E.)
#########################################################################################
class OEBoolean(colander.Boolean):
    """
    Custom boolean type which is more strict.

    The default deserialization casts every value to a Boolean. This is not what we want.
    courtesy of cahytinne
    """

    def deserialize(self, node, cstruct):

        cstruct = {"true": True, "false": False}.get(str(cstruct).lower(), cstruct)

        if cstruct is None:
            cstruct = colander.null  # handling this depends on being required or not

        elif cstruct == colander.null:
            pass

        elif not type(cstruct) is bool:
            raise colander.Invalid(node, "{0} is not a valid Boolean.".format(cstruct))

        result = super(OEBoolean, self).deserialize(node, cstruct)
        return result


class OEInteger(colander.Integer):
    """
    Custom integer type which is more strict.

    The default deserialization casts floats to ints. This is not what we want.
    see also: https://github.com/Pylons/colander/issues/292
    """

    def __init__(self):
        super(OEInteger, self).__init__(strict=True)


#########################################################################################
# SCHEMA NODES (FOR O.E)
#########################################################################################
class OEIDSchemaNode(colander.SchemaNode):
    schema_type = OEInteger

    def validator(self, node, cstruct):
        if cstruct < 0:
            raise colander.Invalid(node, "Geen geldig ID ({}< 0)".format(cstruct))


class OEEmailSchemaNode(colander.SchemaNode):
    schema_type = colander.String
    validator = colander.All(
        colander.Length(max=255), colander.Email("Dit is geen geldig emailadres.")
    )


class OEStringSchemaNode(colander.SchemaNode):
    schema_type = colander.String


class OEDateTimeSchemaNode(colander.SchemaNode):
    schema_type = colander.DateTime


class OEBoolSchemaNode(colander.SchemaNode):
    schema_type = OEBoolean


class UrlSchemaNode(colander.SchemaNode):
    schema_type = colander.String
    title = "Url"

    def validator(self, node, cstruct):
        if len(cstruct) > 2083:
            raise colander.Invalid(node, "URI is too long (max=2083).", cstruct)
        if not rfc3987.match(cstruct, rule="URI"):
            raise colander.Invalid(node, "{0} is geen geldige URL.".format(cstruct))

        try:
            res = requests.get(cstruct, headers={"Accept": "application/json"})
        except requests.ConnectionError:
            raise colander.Invalid(node, "URL bestaat niet.", cstruct)
        if res is not None and res.status_code == 404:
            raise colander.Invalid(node, "URL bestaat niet.", cstruct)


class UriSchemaNode(colander.SchemaNode):
    """
    URL validator.

    rfc3987 is used to check if a URL is correct (https://pypi.python.org/pypi/rfc3987/)
    If ``msg`` is supplied,
    it will be the error message to be used when raising :exc:`colander.Invalid`;
    otherwise, defaults to 'Invalid URL'.
    """

    schema_type = colander.String
    title = "Erfgoed id Uri"

    def validator(self, node, cstruct):
        if len(cstruct) > 100:
            raise colander.Invalid(node, "URI is too long (max=100).", cstruct)
        if not rfc3987.match(cstruct, rule="URI"):
            raise colander.Invalid(node, "{0} is geen geldige URI.".format(cstruct))


class ExternalUriSchemaNode(UriSchemaNode):
    schema_type = colander.String

    def validator(self, node, cstruct):
        super(ExternalUriSchemaNode, self).validator(node, cstruct)

        try:
            res = requests.get(cstruct, headers={"Accept": "application/json"})
        except requests.ConnectionError:
            raise colander.Invalid(node, "URI bestaat niet.", cstruct)
        if res is not None and res.status_code == 404:
            raise colander.Invalid(node, "URI bestaat niet.", cstruct)



# -*- coding: utf-8 -*-
import json
import logging

import requests
from requests import HTTPError
from zope.interface import Interface
from zope.interface import implementer

from oe_utils.data.data_transfer_objects import ResultDTO

log = logging.getLogger(__name__)
NOT_SET = object()


class BulkException(Exception):
    pass


def load_searchquery_parameters(query_params, settings, user_acls=None):
    """
    Create a query for the searchengine based on the provided params.

    :param query_params: the request params for the search action
    :param settings: settings dictionary
    :param user_acls: an acces list for role based filtering
    :returns: Een :class:`dict` query object for the searchengine
    """
    q = {"match_all": {}}
    return q


def default_mapper(result, settings):
    if "hits" in result:
        result = [r["_source"] for r in result["hits"]["hits"]]
        return result
    else:
        return []


class ISearchEngine(Interface):
    def bulk_add_to_index(
        self, system_token, object_type, data, id_field="id", batch_size=5000
    ):
        """Add an iterable of data to the elasticsearch index in bulk."""

    def add_to_index(system_token, object_type, object_id, object_data):
        """Add an object to the index with a specific type."""

    def remove_from_index(system_token, object_type, object_id):
        """Remove an object from the index."""

    def remove_from_index_by_query(system_token, object_field, object_value):
        """Remove an object from the index by query."""

    def query(
        system_token,
        object_type=None,
        query_params=None,
        sort=None,
        result_range=None,
        mapper=default_mapper,
        load_searchquery_param_func=load_searchquery_parameters,
        aggregations=None,
        settings=None,
        user_acls=None,
        min_score=None,
        principals=NOT_SET,
        source=None,
    ):
        """Execute a query on the search engine."""

    def remove_index(system_token):
        """Remove the index."""

    def create_index(system_token, data=None):
        """Create the index."""

    def add_type_mapping(object_type, object_type_mapping, system_token):
        """Add the mapping for specific type."""


@implementer(ISearchEngine)
class SearchEngine(object):
    def __init__(self, baseurl, index_name, version="1"):
        self.baseurl = baseurl
        self.index_name = index_name
        self.version = version
        self.content_header = (
            {"Content-Type": "application/json"} if self.version == "6" else {}
        )

    def assert_response_ok(self, response):
        try:
            response.raise_for_status()
        except HTTPError:
            log.error(response.text)
            raise

    def add_to_index(self, system_token, object_type, object_id, object_data):
        """Add an object to the index with a specific type."""
        if object_type and object_type != "_doc":
            log.info(
                "The use of object types is discouraged by Elasticsearch "
                "and will be depecreated in future versions: "
                "https://www.elastic.co/guide/en/elasticsearch/reference/current/removal-of-types.html "  # noqa
            )
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        headers.update(self.content_header)
        res = requests.put(
            self.baseurl
            + "/"
            + self.index_name
            + "/"
            + object_type
            + "/"
            + str(object_id),
            object_data,
            headers=headers,
        )
        self.assert_response_ok(res)

    def bulk_add_to_index(
        self, system_token, object_type, data, id_field="id", batch_size=5000
    ):
        """
        Add an iterable of data to the elasticsearch index in bulk.

        The _id being set at elasticsearch is taken from the "id_field" in
        each json dict. If the key is not found, no _id will be set and
        Elasticsearch will generate one.

        :param system_token: OpenAmSSOID header for request to Elasticsearch
        :param object_type: The type to add the index to.
        :param data: An iterable of json dicts to add to the index.
        :param id_field: The field in the data which should become the _id in
                         Elasticsearch
        :param batch_size: Maximum amount of items to send per batch.
        :return:
        """
        bulk_url = self.baseurl + "/" + self.index_name + "/" + object_type + "/_bulk"
        batch_data = ""
        for i, item in enumerate(data, start=1):
            action = {"index": {}}
            if id_field in item:
                action["index"]["_id"] = item[id_field]
            batch_data += json.dumps(action) + "\n" + json.dumps(item) + "\n"
            if i % batch_size == 0:
                self._send_batch(system_token, bulk_url, batch_data)
                batch_data = ""
        if batch_data:
            self._send_batch(system_token, bulk_url, batch_data)

    def _send_batch(self, system_token, bulk_url, batch_data):
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        if self.version not in ("1", "2", "3", "4"):
            headers["Content-Type"] = "application/x-ndjson"
        res = requests.put(bulk_url, data=batch_data, headers=headers)
        self.assert_response_ok(res)
        if res.json().get("errors"):
            raise BulkException(
                "Result from bulk operation contains errors: " + res.text
            )

    def remove_from_index(self, system_token, object_type, object_id):
        """Remove an object from the index."""
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        res = requests.delete(
            self.baseurl
            + "/"
            + self.index_name
            + "/"
            + object_type
            + "/"
            + str(object_id),
            headers=headers,
        )
        self.assert_response_ok(res)

    def remove_from_index_by_query(self, system_token, object_field, object_value):
        """Remove an object from the index by query."""
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        if self.version == "1":
            res = requests.delete(
                self.baseurl
                + "/"
                + self.index_name
                + "/_query?q="
                + object_field
                + ":"
                + str(object_value),
                headers=headers,
            )
        else:
            headers.update(self.content_header)
            delete_query = {"query": {"match": {object_field: object_value}}}
            res = requests.post(
                self.baseurl + "/" + self.index_name + "/_delete_by_query",
                data=json.dumps(delete_query),
                headers=headers,
            )
        self.assert_response_ok(res)

    def build_query(
        self,
        query_params=None,
        sort=None,
        result_range=None,
        load_searchquery_param_func=load_searchquery_parameters,
        aggregations=None,
        settings=None,
        user_acls=None,
        min_score=None,
        source=None,
    ):
        """
        Build the elastic search query and the result range params.

        :param self:
        :param query_params: query params passed to load_searchquery_param_func
           and source if callable.
        :param sort: sort value of the query.
        :param result_range: oe_utils Range object to set `from` and `size`
           parameters on the query.
        :param load_searchquery_param_func: callable which takes query_params,
           settings and user_acls=user_acls as parameter. This callable's
           return value will be put in the "query" key of the query.
        :param aggregations: aggregations dict, put directly into the
           "aggregations" key of the query
        :param settings: settings passed to load_searchquery_param_func,
           source if callable and mapper
        :param user_acls: passed to the load_searchquery_param_func and source
           if callable.
        :param min_score: min_score value of the query
        :param source: If callable, acts like load_searchquery_param_func and
           this will be called during the query exection with parameters:
           query_params, settings, user_acls=user_acls.
           The return value will be set as "_source" in the query.
           If not callable this is put directly into the "_source" of the query
        :return: dict that contains the full query and the result range params
        """
        query = load_searchquery_param_func(query_params, settings, user_acls=user_acls)
        if not sort:
            sort = ["_score"]
        params = {}
        if result_range:
            params["size"] = result_range.get_page_size()
            params["from"] = result_range.start
        data = {"query": query, "sort": sort}
        if min_score:
            data["min_score"] = min_score
        if aggregations:
            data["aggregations"] = aggregations
        if source is not None:
            try:
                source = source(query_params, settings, user_acls=user_acls)
            except TypeError:
                pass
            data["_source"] = source

        return {"es_query": data, "params": params}

    def run_query(
        self,
        system_token,
        es_query,
        params,
        object_type=None,
        mapper=default_mapper,
        settings=None,
        principals=NOT_SET,
    ):
        """
        Execute a query on the search engine.

        :param system_token: system token.
        :param es_query: the full elastic search query
        :param params: request paramteres containing the result range
        :param object_type: type in index. When passed, the query will be
           executed to <es>/index/object_type/_search, when None, the query is
           send to <es>/index/_search
        :param mapper: callable used to process the response into data. It will
           receive the ES result, the settings and, if provided, principals as
           parameters. The return result will be put in the `.data` property
           of the response.
        :param settings: settings passed to load_searchquery_param_func,
           source if callable and mapper
        :param principals: this is a parameter given to the mapper function if
           it is not None, otherwise this is not passed.
        :return: ResultDTO with data, total, and aggregations.
        """
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        headers.update(self.content_header)
        search_url = self.baseurl + "/" + self.index_name
        # if no object_type assume full index search
        search_url += "/" + object_type + "/_search" if object_type else "/_search"
        res = requests.post(
            search_url, data=json.dumps(es_query), params=params, headers=headers
        )
        self.assert_response_ok(res)
        result = json.loads(res.text)
        mapper_args = [result, settings]
        if principals != NOT_SET:
            mapper_args.append(principals)
        return ResultDTO(
            mapper(*mapper_args),
            result["hits"]["total"] if "hits" in result else 0,
            result["aggregations"] if "aggregations" in result else None,
        )

    def query(
        self,
        system_token,
        object_type=None,
        query_params=None,
        sort=None,
        result_range=None,
        mapper=default_mapper,
        load_searchquery_param_func=load_searchquery_parameters,
        aggregations=None,
        settings=None,
        user_acls=None,
        min_score=None,
        principals=NOT_SET,
        source=None,
    ):
        """
        Execute a query on the search engine.

        :param system_token: system token.
        :param object_type: type in index. When passed, the query will be
           executed to <es>/index/object_type/_search, when None, the query is
           send to <es>/index/_search
        :param query_params: query params passed to load_searchquery_param_func
           and source if callable.
        :param sort: sort value of the query.
        :param result_range: oe_utils Range object to set `from` and `size`
           parameters on the query.
        :param mapper: callable used to process the response into data. It will
           receive the ES result, the settings and, if provided, principals as
           parameters. The return result will be put in the `.data` property
           of the response.
        :param load_searchquery_param_func: callable which takes query_params,
           settings and user_acls=user_acls as parameter. This callable's
           return value will be put in the "query" key of the query.
        :param aggregations: aggregations dict, put directly into the
           "aggregations" key of the query
        :param settings: settings passed to load_searchquery_param_func,
           source if callable and mapper
        :param user_acls: passed to the load_searchquery_param_func and source
           if callable.
        :param min_score: min_score value of the query
        :param principals: this is a parameter given to the mapper function if
           it is not None, otherwise this is not passed.
        :param source: If callable, acts like load_searchquery_param_func and
           this will be called during the query exection with parameters:
           query_params, settings, user_acls=user_acls.
           The return value will be set as "_source" in the query.
           If not callable this is put directly into the "_source" of the query
        :return: ResultDTO with data, total, and aggregations.
        """
        query_build = self.build_query(
            query_params=query_params,
            sort=sort,
            result_range=result_range,
            load_searchquery_param_func=load_searchquery_param_func,
            aggregations=aggregations,
            settings=settings,
            user_acls=user_acls,
            min_score=min_score,
            source=source,
        )
        return self.run_query(
            system_token,
            es_query=query_build["es_query"],
            params=query_build["params"],
            object_type=object_type,
            mapper=mapper,
            settings=settings,
            principals=principals,
        )

    def remove_index(self, system_token):
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        res = requests.head(self.baseurl + "/" + self.index_name, headers=headers)
        if res.status_code < 400:  # otherwise assume index doens't exists
            res = requests.delete(self.baseurl + "/" + self.index_name, headers=headers)
            self.assert_response_ok(res)

    def create_index(self, system_token, data):
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        headers.update(self.content_header)
        res = requests.put(
            self.baseurl + "/" + self.index_name, data=json.dumps(data), headers=headers
        )
        self.assert_response_ok(res)

    def add_type_mapping(self, object_type, object_type_mapping, system_token):
        headers = {"OpenAmSSOID": system_token} if system_token else {}
        headers.update(self.content_header)
        res = requests.put(
            self.baseurl + "/" + self.index_name + "/_mapping/" + object_type,
            data=json.dumps(object_type_mapping),
            headers=headers,
        )
        self.assert_response_ok(res)



# -*- coding: utf-8 -*-
from collections import Sequence


def conditional_http_tween_factory(handler, registry):
    """Tween that adds ETag headers and enables conditional responses."""
    settings = registry.settings if hasattr(registry, "settings") else {}
    not_cacheble_list = []
    if "not.cachable.list" in settings:
        not_cacheble_list = settings.get("not.cachable.list").split()

    def conditional_http_tween(request):
        response = handler(request)

        if request.path not in not_cacheble_list:

            # If the Last-Modified header has been set, we want to enable the
            # conditional response processing.
            if response.last_modified is not None:
                response.conditional_response = True

            # We want to only enable the conditional machinery if either we
            # were given an explicit ETag header by the view or we have a
            # buffered response and can generate the ETag header ourself.
            if response.etag is not None:
                response.conditional_response = True
            elif (
                isinstance(response.app_iter, Sequence) and len(response.app_iter) == 1
            ) and response.body is not None:
                response.conditional_response = True
                response.md5_etag()

        return response

    return conditional_http_tween



# -*- coding: utf-8 -*-
import functools
import hashlib
import json
import logging
from datetime import datetime

from pyramid import renderers
from pytz import timezone

log = logging.getLogger(__name__)


def _action_from_request(request):
    if request.method == "POST":
        return "aanmaken"
    elif request.method == "PUT":
        return "bewerken"
    elif request.method == "DELETE":
        return "verwijderen"
    elif request.method == "GET":
        return "opvragen"
    else:
        return "onbekend"


def _get_versie_hash(wijziging):
    timestamp = datetime.now(tz=timezone("CET"))
    inputversie = "{0}{1}{2}".format(
        timestamp, wijziging.resource_object_id, wijziging.updated_by
    )
    versie = hashlib.sha256(bytearray(inputversie, "utf-8")).hexdigest()
    return versie


def _get_id_from_result(r_id, result, kwarg):
    if kwarg.get("result_id_key"):
        return int(result[kwarg.get("result_id_key")])
    return result["id"] if not r_id else int(r_id)


def audit(**kwargs):
    """Audit an operation."""
    def wrap(fn):
        @functools.wraps(fn)
        def advice(parent_object, *args, **kw):
            request = parent_object.request
            wijziging = request.audit_manager.create_revision()

            result = fn(parent_object, *args, **kw)

            if (
                hasattr(request, "user")
                and request.user is not None
                and "actor" in request.user
            ):
                actor = request.user["actor"]
                attributes = request.user["attributes"]
                wijziging.updated_by = actor.get("uri", None)
                if actor.get("uri") == actor.get("instantie_actor_uri"):
                    wijziging.updated_by_omschrijving = (
                        attributes.get("displayname")
                        or attributes.get("mail")
                        or actor.get("omschrijving")
                    )
                else:
                    wijziging.updated_by_omschrijving = actor.get("omschrijving")
            else:
                wijziging.updated_by = "publiek"
                wijziging.updated_by_omschrijving = "publiek"

            r_id = request.matchdict.get("id")
            wijziging.resource_object_id = r_id
            if result is not None:
                try:
                    renderer_name = request.registry.settings.get(
                        "audit.pyramid.json.renderer", "jsonrenderer"
                    )
                    json_string = renderers.render(
                        renderer_name, result, request=request
                    )
                    result_object_json = json.loads(json_string)
                    wijziging.resource_object_json = result_object_json
                    wijziging.resource_object_id = _get_id_from_result(
                        r_id, result_object_json, kwargs
                    )
                except Exception as e:
                    log.exception(e)

            wijziging.versie = _get_versie_hash(wijziging)
            wijziging.actie = (
                kwargs.get("actie")
                if kwargs.get("actie")
                else _action_from_request(request)
            )

            request.audit_manager.save(wijziging)

            return result

        return advice

    return wrap


def audit_with_request(**kwargs):
    """Audit an operation with a request as input variable."""
    def wrap(fn):
        @audit(**kwargs)
        def operation(parent_object, *args, **kw):
            return fn(parent_object.request, *args, **kw)

        @functools.wraps(fn)
        def advice_with_request(the_request, *args, **kw):
            class ParentObject:
                request = the_request

            return operation(ParentObject(), *args, **kw)

        return advice_with_request

    return wrap



# -*- coding: utf-8 -*-
import json
from copy import copy


class AdapterJSONEncoder(json.JSONEncoder):
    """
    A JSONEncoder which has a dict of adapters to map classes to json.

    Each class has its own custom 'adapter' method which can turn an instance
    of this class into a dict.
    By default these adapter methods take only the object as parameter, but
    extra kwargs can optionally be added with the 'extra_kwargs' parameter.
    """

    def __init__(self, adapters, extra_kwargs, json_dump_kwargs):
        super(AdapterJSONEncoder, self).__init__(**json_dump_kwargs)
        self.adapters = adapters
        self.extra_kwargs = extra_kwargs

    def default(self, obj):
        object_type = type(obj)
        try:
            return self.adapters[object_type](obj, **self.extra_kwargs)
        except KeyError:
            return json.JSONEncoder.default(self, obj)


class CustomizableJSONifier(object):

    def __init__(self, json_dump_kwargs=None, **adapter_kwargs):
        """
        Create an object to which custom adapters can be registered.

        Once registered, this class can turn custom class instances into json.

        :type json_dump_kwargs: dict
        :param json_dump_kwargs: kwargs which will be passed on to the json
           dump / dumps methods. For example 'indent', 'sort_keys'
        :param adapter_kwargs: Any extra kwarg will be passed to the
           registered adapters.
        """
        super(CustomizableJSONifier, self).__init__()
        self.adapters = {}
        self.default_json_dump_kwargs = json_dump_kwargs or {}
        self.default_adapter_kwargs = adapter_kwargs

    def __call__(self, custom_adapter_kwargs, custom_dump_kwargs=None,
                 *args, **kwargs):
        kwargs.update(custom_dump_kwargs)
        return AdapterJSONEncoder(self.adapters, custom_adapter_kwargs, kwargs)

    def register_adapter(self, class_, adapter=None):
        """Register a method to be used when serializing the class to json."""
        if adapter is None:
            def decorator(fn):
                self.adapters[class_] = fn
                return fn
            return decorator
        else:
            self.adapters[class_] = adapter

    def dumps(self, obj, json_dump_kwargs=None, **adapter_kwargs):
        """
        Turn an object into a json string.

        :param obj: The object to turn into json.
        :type json_dump_kwargs: dict
        :param json_dump_kwargs: kwargs which will be passed on to the json
           dump / dumps methods. For example 'indent', 'sort_keys'. These
           kwargs will take precedence over the ones set on the encoder.
        :param adapter_kwargs: Any extra kwarg will be passed to the
           registered adapters.
        :rtype: str
        :return: a json string.
        """
        custom_adapter_kwargs, custom_dump_kwargs = self._prepare_kwargs(
            adapter_kwargs, json_dump_kwargs
        )
        return json.dumps(obj, cls=self,
                          custom_adapter_kwargs=custom_adapter_kwargs,
                          custom_dump_kwargs=custom_dump_kwargs)

    def to_dict(self, obj, json_dump_kwargs=None, **adapter_kwargs):
        """
        Turn an object into a dict.

        This method simply uses dumps to turn the object into a string and
        then loads it back with json.loads.

        :param obj: The object to turn into json.
        :type json_dump_kwargs: dict
        :param json_dump_kwargs: kwargs which will be passed on to the json
           dump / dumps methods. For example 'indent', 'sort_keys'. These
           kwargs will take precedence over the ones set on the encoder.
        :param adapter_kwargs: Any extra kwarg will be passed to the
           registered adapters.
        :return: a json object. Probably dict or list.
        """
        json_string = self.dumps(obj, json_dump_kwargs=json_dump_kwargs,
                                 **adapter_kwargs)
        return json.loads(json_string)

    def dump(self, obj, fp, json_dump_kwargs=None, **adapter_kwargs):
        """Turn an object into a json string written to a file.

        :param fp: a file-like object. (has .write() method)
        :param obj: The object to turn into json.
        :type json_dump_kwargs: dict
        :param json_dump_kwargs: kwargs which will be passed on to the json
           dump / dumps methods. For example 'indent', 'sort_keys'. These
           kwargs will take precedence over the ones set on the encoder.
        :param adapter_kwargs: Any extra kwarg will be passed to the
           registered adapters.
        :rtype: str
        :return: a json string.
        """
        custom_adapter_kwargs, custom_dump_kwargs = self._prepare_kwargs(
            adapter_kwargs, json_dump_kwargs
        )
        return json.dump(obj, fp, cls=self,
                         custom_adapter_kwargs=custom_adapter_kwargs,
                         custom_dump_kwargs=custom_dump_kwargs)

    def _prepare_kwargs(self, adapter_kwargs, json_dump_kwargs):
        custom_dump_kwargs = copy(self.default_json_dump_kwargs)
        if json_dump_kwargs is not None:
            custom_dump_kwargs.update(json_dump_kwargs)
        custom_adapter_kwargs = copy(self.default_adapter_kwargs)
        if adapter_kwargs is not None:
            custom_adapter_kwargs.update(adapter_kwargs)
        return custom_adapter_kwargs, custom_dump_kwargs


_global_encoder = CustomizableJSONifier()
register_adapter = _global_encoder.register_adapter
dump = _global_encoder.dump
dumps = _global_encoder.dumps
to_dict = _global_encoder.to_dict



# -*- coding: utf-8 -*-
import math
import re

from past.builtins import basestring
from webob.multidict import MultiDict

try:
    # python 3
    from urllib.parse import urlencode
except ImportError:
    # python 2
    from urllib import urlencode


DEFAULT_MAX_END = "default"


class RangeParseException(Exception):
    """Used for handling errors during range parsing."""


class Range(object):
    def __init__(self, start, end):
        super(Range, self).__init__()
        self.start = start
        self.end = end


class ItemRange(Range):
    def __init__(self, start, end, page=1):
        super(ItemRange, self).__init__(start, end)
        self.page = page

    def get_page_size(self):
        return self.end - self.start + 1

    def content_range(self, length):
        """
        Return a string representation of a range.

        :param length: the total number of results
        :return: range header string
        """
        end = self.end if self.end <= length else length - 1
        return "items %d-%d/%d" % (self.start, end, length)

    @classmethod
    def parse(cls, request, default_start=0, default_end=9, max_end=DEFAULT_MAX_END):
        """
        Parse the range headers into a range object.

        When there are no range headers, check for a page 'pagina' parameter,
        otherwise use the defaults defaults.

        :param request: a request object
        :param default_start: default start for  paging (optional, default is 0)
        :param default_end: default end for paging (optional, default is 9)
        :param max_end: maximum end for paging (optional, default is 50,
                        if oe.paging.max_end settings is set, this value will be used)
        :return: :class: 'oe_utils.range_parser.Range'
        """
        settings = request.registry.settings
        page_param = settings.get("oe.paging.page.queryparam", "pagina")
        per_page_param = settings.get("oe.paging.per_page.queryparam", "per_pagina")
        max_end_settings = settings.get("oe.paging.max_end", None)

        # handling overriding of max_end by coders.
        if max_end_settings and max_end == DEFAULT_MAX_END:
            max_end = int(max_end_settings)
        # if no max_end, fallback to default value
        if max_end == DEFAULT_MAX_END:
            max_end = 50
        range_header = request.headers.get("Range")
        if range_header and range_header.startswith('items='):
            match = re.match("^items=([0-9]+)-([0-9]+)$", range_header)

            if match:
                start = int(match.group(1))
                end = int(match.group(2))

                if end < start:
                    end = start
                if max_end and end > start + max_end:
                    end = start + max_end
                return cls(start, end)
            else:
                raise RangeParseException(
                    "range header '{}' does not match expected format".format(
                        range_header
                    )
                )
        elif page_param in request.params:
            page = int(request.params.get(page_param))
        else:
            page = 1

        items_per_page = int(
            request.params.get(per_page_param, default_end - default_start + 1)
        )
        # enforcing items_per_page is less than the settings limit.
        if max_end is not None and items_per_page > max_end:
            items_per_page = max_end
        start = default_start + items_per_page * (page - 1)
        end = start + items_per_page - 1
        return cls(start, end, page)

    def set_response_headers(self, request, total_count):
        """
        Set the correct range headers on the response.

        :param request: a request object
        :param total_count: the total number of results
        """
        response = request.response
        response.headerlist.append(
            ("Access-Control-Expose-Headers", "Content-Range, X-Content-Range")
        )
        response.accept_ranges = "items"
        if total_count is None:
            raise RangeParseException("Provided length value is null")
        if total_count > 0:
            response.content_range = self.content_range(total_count)

        self.set_link_headers(request, total_count)

    def set_link_headers(self, request, total_count):
        """
        Set Link headers on the response.

        When the Range header is present in the request no Link headers will
        be added.
        4 links will be added: first, prev, next, last.
        If the current page is already the first page, the prev link will
        not be present.
        If the current page is already  the last page, the next link will
        not be present.

        :param request: A request object
        :param total_count: The total amount of items available before paging
        """
        response = request.response
        if request.headers.get("Range"):
            # Don't set the Link headers when custom ranges were used.
            return

        settings = request.registry.settings
        page_param = settings.get("oe.paging.page.queryparam", "pagina")
        per_page_param = settings.get("oe.paging.per_page.queryparam", "per_pagina")
        url = request.path_url
        # params by default is read-only so we put it in new editable MultiDict
        queryparams = MultiDict(request.params)
        page_size = self.get_page_size()
        current_page = self.start // page_size + 1
        queryparams[per_page_param] = page_size
        links = {"first": 1, "last": int(math.ceil(float(total_count) / page_size))}
        if current_page != links["first"]:
            links["prev"] = current_page - 1
        if current_page != links["last"]:
            links["next"] = current_page + 1

        response.headers["Link"] = self._make_link_headers(
            links, page_param, queryparams, url
        )

    def _make_link_headers(self, links, page_param, queryparams, url):
        link_headers = []
        queryparams = MultiDict(
            [
                (key, val.encode("utf8") if isinstance(val, basestring) else val)
                for key, val in queryparams.items()
            ]
        )
        for name, page in links.items():
            queryparams[page_param] = page
            link_headers.append(
                '<{url}?{queryparams}>; rel="{name}"'.format(
                    url=url, queryparams=urlencode(queryparams), name=name
                )
            )
        return ", ".join(link_headers)



# -*- coding: utf-8 -*-
import argparse


def create_indexing_argument_parser(choices):
    parser = argparse.ArgumentParser(
        description="Process some command line arguments.", prog="reindexes"
    )
    parser.add_argument("--offset", type=int, help="starting point of indexing")
    parser.add_argument("--limit", type=int, help="endpoint of indexing")
    parser.add_argument(
        "--index", type=str, choices=choices, help="catagory of objects to be indexed"
    )
    parser.add_argument("--id", type=int, help="id of a given item to be indexed")
    return parser



# -*- coding: utf-8 -*-
import logging
import sys

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import forbidden_view_config
from pyramid.view import notfound_view_config
from pyramid.view import view_config
from sqlalchemy.orm.exc import NoResultFound

from oe_utils.validation import ValidationFailure

log = logging.getLogger(__name__)


@view_config(context=Exception, renderer="json", accept="application/json")
def internal_server_error(exception, request):
    log.error(str(exception), exc_info=sys.exc_info())
    request.response.status_int = 500
    return {
        "message": "Er ging iets fout in de server. Onze excuses. "
        "Stel je fouten vast of heb je een vraag? Mail dan naar ict@onroerenderfgoed.be"
    }


@notfound_view_config(renderer="json", accept="application/json")
def not_found(request):
    request.response.status_int = 404
    return {"message": "De door u gevraagde resource kon niet gevonden worden."}


@view_config(context=NoResultFound, renderer="json")
def failed_no_result_found(exc, request):
    request.response.status_int = 404
    return {"message": "De door u gevraagde resource kon niet gevonden worden."}


@view_config(context=HTTPBadRequest, renderer="json", accept="application/json")
def json_bad_request(exc, request):
    request.response.status_int = 400
    return {"message": exc.message}


@view_config(context=ValidationFailure, renderer="json")
def failed_validation(exc, request):
    log.debug(exc.msg)
    log.debug(exc.errors)
    request.response.status_int = 400
    formated_errors = []
    for node in exc.errors:
        formated_errors.append(
            " ".join(list(reversed(node.split("."))))
            .capitalize()
            .replace("_", " ")
            .replace("body", "")
            + ": "
            + exc.errors[node]
            if node != ""
            else exc.errors[node]
        )
    return {"message": exc.msg, "errors": formated_errors}


@forbidden_view_config(accept="application/json", renderer="json")
def forbidden_view(exc, request):
    log.debug("FORBIDDEN anything")
    log.debug(exc)
    log.debug(exc.message)
    log.debug(exc.result)
    # do not allow a user to login if they are already logged in
    if request.authenticated_userid:
        err = {
            "message": "U bent niet gemachtigd om deze actie uit te voeren.",
            "errors": [],
        }
        if exc.result.msg:
            err["errors"].append(exc.result.msg)
        request.response.status_int = 403
        return err



# -*- coding: utf-8 -*-


def get_oe_email_footer_html():
    return """
    <div style="float:left; width:100%;">
        Onroerend Erfgoed<br/>
        Herman Teirlinckgebouw  |  Havenlaan 88 bus 5  |  1000 Brussel <br/>
        <div style="width:100%">
            <div style="float:left; width:100%">
                <a href='https://www.onroerenderfgoed.be'>www.onroerenderfgoed.be</a>
            </div>
            <div style="float:left; width:100%">
                <img src="https://dgen.onroerenderfgoed.be/static/logo_vlaanderen_footer.svg" height="100">
            </div>
        </div>
    </div>
    """  # noqa E501


def get_oe_email_footer_plain():
    return (
        "Onroerend Erfgoed\n"
        "Herman Teirlinckgebouw  |  Havenlaan 88 bus 5  |  1000 Brussel\n"
        "www.onroerenderfgoed.be"
    )



# -*- coding: utf-8 -*-
try:
    import configparser
except ImportError:  # pragma: no cover
    import ConfigParser as configparser  # pragma: no cover
import logging
import requests


log = logging.getLogger(__name__)


def archive_entries(feed_list, session, system_token):
    for feed_obj in feed_list:
        atom_feed_manager = feed_obj['atom_feed_manager'](session, feed_obj['atom_feed_folder'])
        archive_feed_url = feed_obj['archive_feed_url']
        max_entries = feed_obj['max_entries']

        current_feed = atom_feed_manager.current_feed
        entries_count = len(current_feed.entries)
        log.debug("Number of entries: {}".format(entries_count))
        if entries_count >= max_entries:
            log.debug("Archive current feed")
            archive_feed = requests.post(
                archive_feed_url, headers={"OpenAmSSOID": system_token, "Accept": "Application/json"})
            if archive_feed.status_code != 200:
                log.error("Failed to archive on url {}".format(archive_feed_url))
                archive_feed.raise_for_status()



# -*- coding: utf-8 -*-
import abc

from feedgen.feed import FeedGenerator


class AbstractAtomFeedView:

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_current_atom_feed(self):
        """
        Retrieve the current atom feed.

        :return: JSON representation of the feed
        """

    @abc.abstractmethod
    def get_atom_feed(self):
        """
        Retrieve a feed identified by its id, identical to the source.

        :return: JSON representation of the feed
        """

    @abc.abstractmethod
    def get_atom_feed_entry(self):
        """
        Retrieve a feed entry identified by its id, identical to the source.

        :return: JSON representation of the feed entry
        """


class AtomFeedView(AbstractAtomFeedView):
    """
    Class with pyramid support.

    This is an implementation of the :class:`AbstractAtomFeedView` that adds a
    generic methods in a pylons pyramid application
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self, request, atom_feed_manager, get_atom_feed_url, generate_atom_feed=None
    ):
        """
        Functions that can be used in a pylons pyramid application.

        :param request:
        :param atom_feed_manager:
        object of class oe_daemonutils.data.data_manager.AtomFeedManager
        :param get_atom_feed_url:
        the route name of the pyramid application to retrieve the atom feed
        """
        self.request = request
        self.atom_feed_manager = atom_feed_manager
        self.get_atom_feed_url = get_atom_feed_url
        self.generate_atom_feed = (
            generate_atom_feed if generate_atom_feed else self._generate_atom_feed
        )

    def get_current_atom_feed(self):
        return self.request.route_url(
            self.get_atom_feed_url, id=self.atom_feed_manager.current_feed_id
        )

    def get_atom_feed(self):
        feed_id = int(self.request.matchdict["id"])
        return self.generate_atom_feed(feed_id).atom_str(pretty=True)

    def get_atom_feed_entry(self):
        return self.atom_feed_manager.get_atom_feed_entry(
            int(self.request.matchdict["id"])
        )

    def link_to_sibling(self, feed_id, sibling_type, atom_feed):
        """
        Adding previous or next links to the given feed.

        self._link_to_sibling(feed, 'previous', atom_feed)
        self._link_to_sibling(feed, 'next', atom_feed)

        :param feed_id: a feed_id
        :param sibling_type: 'previous' or 'next'
        :param atom_feed: an atom feed like `feedgen.feed.FeedGenerator`
        """
        sibling = self.atom_feed_manager.get_sibling_id(feed_id, sibling_type)
        if sibling:
            rel = "prev-archive" if sibling_type == "previous" else "next-archive"
            atom_feed.link(
                href=self.request.route_url(self.get_atom_feed_url, id=sibling),
                rel=rel,
            )

    def init_atom_feed(self, feed_id):
        """
        Initialise an atom feed `feedgen.feed.FeedGenerator`.

        :param feed_id: feed_id
        :return: an atom feed `feedgen.feed.FeedGenerator`
        """
        atom_feed = FeedGenerator()
        atom_feed.id(id=self.request.route_url(self.get_atom_feed_url, id=feed_id))
        atom_feed.link(
            href=self.request.route_url(self.get_atom_feed_url, id=feed_id), rel="self"
        )
        atom_feed.language("nl-BE")
        self.link_to_sibling(feed_id, "previous", atom_feed)
        self.link_to_sibling(feed_id, "next", atom_feed)
        return atom_feed

    def _generate_atom_feed(self, feed_id):
        """
        Return a feed like `feedgen.feed.FeedGenerator`.

        The function can be overwritten when used in other applications.

        :param feed_id: feed_id
        :return: an atom feed `feedgen.feed.FeedGenerator`
        """
        atom_feed = self.init_atom_feed(feed_id)
        atom_feed.title("Feed")
        return atom_feed



# -*- coding: utf8 -*-
import datetime
import math
import os


def get_last_modified_date(filename):
    """
    Get the last modified date of a given file.

    :param filename: string: pathname of a file
    :return: Date
    """
    if os.path.isfile(filename):
        t = os.path.getmtime(filename)
        return datetime.date.fromtimestamp(t).strftime("%d/%m/%Y")
    return None


def get_file_size(filename):
    """
    Get the file size of a given file.

    :param filename: string: pathname of a file
    :return: human readable filesize
    """
    if os.path.isfile(filename):
        return convert_size(os.path.getsize(filename))
    return None


def convert_size(size_bytes):
    """
    Transform bytesize to a human readable filesize.

    :param size_bytes: bytesize
    :return: human readable filesize
    """
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])



# -*- coding: utf-8 -*-
from math import ceil

from deprecated import deprecated
from pyramid.httpexceptions import HTTPBadRequest
from webob.multidict import MultiDict


ES_FIELDS = ('_score', '_id', '_type', '_index', '_source')


def validate_params(params, valid_keys, fail_on_invalid_key=True):
    p = MultiDict()
    for k, v in params.items():
        if k in valid_keys:
            p.add(k, v)
        elif fail_on_invalid_key:
            raise HTTPBadRequest("invalid search parameter")
    return p


@deprecated(version="1.5.0", reason="You should use validate_params")
def get_valid_params(params, valid_keys):
    return validate_params(params, valid_keys, fail_on_invalid_key=False)


def parse_sort_string(
    sort, default=None, fulltext=None, tiebreaker=None, **translations
):
    """
    Parse a sort string for use with elasticsearch.

    If no custom default and no sort string: "_score" becomes the default.

    :param sort: the sort string
    :type sort: str, optional
    :param default: optional custom default sort parameter
    :type default: list, optional
    :param fulltext:  maps fulltext field to "_score"
    :type: fulltext: str, optional
    :param tiebreaker: optional tiebreaker to break equal sort scores
    :type tiebreaker: list, optional
    :param translations: **kwarg that translates sort field string to correct ES field
                          e.g "name" = "name.raw"
    :return: ES ready sort list
    """
    added_field_names = set()
    sortlist = []
    if sort:
        sort_values = sort.rsplit(",")
        for sort_value in sort_values:
            sort_value = sort_value.strip()
            order = "desc" if sort_value[0:1] == "-" else "asc"
            field = sort_value[1:] if sort_value[0:1] in ["-", "+"] else sort_value
            if field == fulltext:
                added_field_names.add("_score")
                sortlist.append("_score")
                continue

            if translations:
                field = translations.get(field, field)

            added_field_names.add(field)

            field_sort = {
                field: {
                    "order": order
                }
            }
            if field not in ES_FIELDS:
                field_sort[field]['unmapped_type'] = 'string'
                field_sort[field]['missing'] = '_last'
            sortlist.append(field_sort)
    elif default:
        sortlist.extend(default)
    if "_score" not in added_field_names:
        sortlist.append("_score")
    if tiebreaker:
        sortlist.extend(tiebreaker)

    return sortlist


def parse_filter_params(query_params, filterable):
    """
    Parse query_params to a filter params dict.

    Merge multiple values for one key to a list.
    Filter out keys that aren't filterable.

    :param query_params: query params
    :param filterable: list of filterable keys
    :return: dict of filter values
    """
    if query_params is not None:
        filter_params = {}
        for fq in query_params.mixed():
            if fq in filterable:
                filter_params[fq] = query_params.mixed().get(fq)
        return filter_params
    else:
        return {}


class SearchResultPager(object):
    """
    Helper class for paging search results.

    based on http://atlas-core.readthedocs.org/
    en/latest/_modules/flask_sqlalchemy.html Pagination
    """

    def __init__(self, page, per_page, total_count):
        self.page = page
        self.per_page = per_page
        self.total_count = total_count

    @property
    def pages(self):
        return int(ceil(self.total_count / float(self.per_page)))

    @property
    def has_prev(self):
        return self.page > 1

    @property
    def prev(self):
        return self.page - 1 if self.has_prev else None

    @property
    def has_next(self):
        return self.page < self.pages

    @property
    def next(self):
        return self.page + 1 if self.has_next else None

    def iter_pages(self, left_edge=2, left_current=2, right_current=5, right_edge=2):
        """Iterate over the page numbers in the pagination.

        The four parameters control the thresholds how many numbers should
        be produced from the sides.  Skipped page numbers are represented
        as `None`.
        """
        last = 0
        try:
            xrange  # noqa
        except NameError:
            xrange = range
        for num in xrange(1, self.pages + 1):
            if (
                num <= left_edge
                or (self.page - left_current - 1 < num < self.page + right_current)
                or num > self.pages - right_edge
            ):
                if last + 1 != num:
                    yield None
                yield num
                last = num



# -*- coding: utf-8 -*-
import logging

from oe_utils.data.data_managers import AuditManager

log = logging.getLogger(__name__)


def audit_manager(request):
    """
    Initialize and return the audit manager.

    The request must contain a method `request.db` to retrieve the current session.

    :param request: `pyramid.request.Request`
    :return: the audit manager
    """
    session = request.db
    return AuditManager(session)


def includeme(config):
    """
    Configure a `request.audit_managers` method to retrieve the audit manager.

    :param config:
    """
    config.add_request_method(audit_manager, reify=True)



import yaml


def copy_and_replace(file_in, file_out, mapping, **kwargs):
    """Copy a file and replace some placeholders with new values."""
    separator = "@@"
    if "separator" in kwargs:
        separator = kwargs["separator"]
    file_in = open(file_in, "r")
    file_out = open(file_out, "w")
    s = file_in.read()
    for find, replace in mapping:
        find = separator + find + separator
        print(u"Replacing {0} with {1}".format(find, replace))
        s = s.replace(find, replace)
    file_out.write(s)


def get_yaml(file_in):
    yamlfile = ""
    with open(file_in, "r") as stream:
        try:
            yamlfile = yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    return yamlfile


def get_wildcards(file_in):
    wildcards = get_yaml(file_in)
    mapping = []
    for category in wildcards["variables"]:
        categories = wildcards["variables"][category]
        for var in categories:
            a = var
            b = categories[var]
            wtuple = (a, b)
            mapping.append(wtuple)
    return mapping


def append_extension(file_in, file_out, extension):
    a = 0

    with open(file_in, "r") as in_file:
        buf = in_file.readlines()

    with open(file_out, "w") as out_file:
        for line in buf:
            if "egg" in line and a == 0:
                line = line + "\n" + extension
                a += 1
            out_file.write(line)



# -*- coding: utf-8 -*-
# based upon https://gist.github.com/eoghanmurray/1360da8635f0a6a6d528
from sqlalchemy.ext.mutable import Mutable


class MutableList(Mutable, list):
    @classmethod
    def coerce(cls, key, value):
        """Convert plain list to MutableList."""
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __getstate__(self):
        return list(self)

    def __setstate__(self, state):
        self.extend(state)

    def append(self, value):
        list.append(self, value)
        self.changed()

    def extend(self, iterable):
        list.extend(self, iterable)
        self.changed()

    def insert(self, index, item):
        list.insert(self, index, item)
        self.changed()

    def pop(self, index=None):
        if index:
            list.pop(self, index)
        else:
            list.pop(self)
        self.changed()

    def remove(self, item):
        list.remove(self, item)
        self.changed()

    def reverse(self):
        list.reverse(self)
        self.changed()

    def sort(self, **kwargs):
        list.sort(self, **kwargs)
        self.changed()



# -*- coding: utf-8 -*-
import json

import requests
from zope.interface import Interface
from zope.interface import implementer


class IDocumentEngine(Interface):
    def create_documents(self, data, system_token, callback_url=None):
        """Create the documents with optional callback_url."""

    def get_document(self, generation_id, system_token, accept="application/pdf"):
        """
        Get the document.

        Accept is either application/pdf (default) or either text/html
        """


@implementer(IDocumentEngine)
class DocumentEngine(object):
    def __init__(self, baseurl, template_id):
        self.baseurl = baseurl
        self.template_id = template_id

    def create_documents(
        self,
        data,
        system_token,
        callback_url=None,
        callback_method=None,
        template_version=None,
    ):
        """Create the documents with optional callback_url."""
        generation_data = {"template_id": self.template_id, "data": data}
        if template_version is not None:
            generation_data["template_version"] = template_version
        if callback_url is not None:
            generation_data["callback_method"] = (
                callback_method if callback_method is not None else "POST"
            )
            generation_data["callback_url"] = callback_url
        headers = {
            "OpenAmSSOID": system_token,
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        res = requests.post(
            "{0}/generations".format(self.baseurl),
            json=generation_data,
            headers=headers,
        )
        res.raise_for_status()
        return json.loads(res.text)["id"]

    def get_document(self, generation_id, system_token, accept="application/pdf"):
        """
        Get the document.

        Accept is either application/pdf (default) or either text/html.
        """
        headers = {"OpenAmSSOID": system_token, "Accept": accept}
        res = requests.get(
            "{0}/generations/{1}/document".format(self.baseurl, generation_id),
            headers=headers,
        )
        res.raise_for_status()
        return res.content



# -*- coding: utf-8 -*-
"""
This module allows for the validation of actor data.

Data like national ID number and telephone.
"""

import re

import colander
from stdnum.be import vat
from stdnum.exceptions import InvalidFormat
from stdnum.exceptions import InvalidLength


def is_number(s):
    try:
        int(s)
    except ValueError:
        return False
    else:
        return True


class RRNSchemaNode(colander.SchemaNode):
    title = "rrn"
    schema_type = colander.String

    @staticmethod
    def validator(node, rrn):
        """
        Colander validator that checks whether a given value is valid.

        A value is valid when it is a valid Belgian national ID number.
        :raises colander.Invalid: If the value is no valid Belgian national ID number.
        """
        def _valid(_rrn):
            x = 97 - (int(_rrn[:-2]) - (int(_rrn[:-2]) // 97) * 97)
            return int(_rrn[-2:]) == x

        if len(rrn) != 11:
            raise colander.Invalid(
                node, "Een rijksregisternummer moet 11 cijfers lang zijn."
            )
        elif not _valid(rrn):
            valid_rrn = False
            if rrn[:1] == "0" or rrn[:1] == "1":
                rrn = "2" + rrn
                valid_rrn = _valid(rrn)
            if not valid_rrn:
                raise colander.Invalid(node, "Dit is geen correct rijksregisternummer.")


class KBOSchemaNode(colander.SchemaNode):
    title = "kbo"
    schema_type = colander.String

    @staticmethod
    def preparer(value):
        """Edit a value to a value that can be validated as a kbo number."""
        if value is None or value == colander.null:
            return colander.null
        return vat.compact(value)

    @staticmethod
    def validator(node, value):
        """
        Colander validator that checks whether a given value is valid.

        For our purposes , we expect a firm number that
        is composed of nine or ten characters like 2028445291.
        Sometimes a company number is formatted with separation marks like 0.400.378.485.
        Therefore, there is also a : Func:
        ` actoren.validators.kbo_preparer` which transforms such input.
        :raises colander.Invalid: if the value is valid Belgian company number.
        """
        try:
            return vat.validate(value)
        except InvalidLength:
            msg = "Het ondernemingsnummer moet 10 karakters lang zijn."
        except InvalidFormat:
            msg = "Het ondernemingsnummer heeft een fout formaat."
        except Exception:
            msg = "Het ondernemingsnummer is niet geldig."
        raise colander.Invalid(node, msg=msg)


class TelefoonSchemaNode(colander.MappingSchema):
    landcode = colander.SchemaNode(colander.String(), missing="+32")
    nummer = colander.SchemaNode(colander.String(), missing="")

    @staticmethod
    def preparer(telefoon):
        """
        Edit a value so that it can be validated as a phone number.

        This takes the incoming value and :
            Removes all whitespace ( space, tab , newline , ... ) characters
            Removes the following characters: " / - . "
            If no + is present at frond, add the country code
            In short: just add a + at the beginning of the country code.
        """
        if telefoon is None or telefoon == colander.null:
            return colander.null
        if "landcode" in telefoon and telefoon.get("landcode") is not None:
            landcode = telefoon.get("landcode")
            value = re.sub(r"[\s./,-]", "", landcode).lstrip("0")
            telefoon["landcode"] = "+" + value if value[0] != "+" else value
        if "nummer" in telefoon and telefoon.get("nummer") is not None:
            nummer = telefoon.get("nummer")
            value = re.sub(r"[\s./,-]", "", nummer).lstrip("0")
            telefoon["nummer"] = value
        return telefoon

    @staticmethod
    def validator(node, telefoon):
        """
        Validate a phone number.

        A valid international phone number looks like this: + .
        It is up to 15 digits long . The country code consists of 1 to 4 digits.
        The actual number may even 15 - cells ( country code) figures cover .
        We're going to keep the phone numbers
        including a + to indicate the international call prefix.
        A valid number always begins with a + .
        The shortest valid number that we want to keep is: +11 ( 3 characters).
        The longest : +123123456789123 (16 characters).
        Additional validation (after preparation ) :
        If a number starts with +32 (this is a Belgian number),
        it must then follow eight or nine digits .
        The first decimal after +32 may not be 0 .
        """
        if telefoon.get("nummer", "") != "":
            msg = "Invalid phone number"
            landcode = telefoon.get("landcode", "")
            nummer = telefoon.get("nummer", "")
            if (
                landcode[0] != "+"
                or len(landcode[1:]) > 4
                or not is_number(landcode[1:])
            ):
                raise colander.Invalid(
                    node,
                    msg + ": Een geldige landcode begint met een + gevolgd door "
                    "maximaal 4 cijfers",
                )
            if (
                landcode[0] != "+"
                or len(landcode[1:]) + len(nummer) > 15
                or not is_number(nummer)
            ):
                raise colander.Invalid(
                    node,
                    msg + ": Een geldige nummer begint met een + gevolgd door "
                    "maximaal 15 cijfers",
                )
            # validatie van Belgische telefoonnummers
            if landcode == "+32":
                if len(nummer) not in [8, 9]:
                    raise colander.Invalid(
                        node, msg + ": Na +32 moeten er 8 of 9 cijfers volgen"
                    )



"""
Dit bestand bevat een OE Base Model dat in alle OE toepassingen kan worden gebruikt. Het voordeel van een generieke
Base is dat we op deze werkwijze per toepassing 1 zelfde Base kunnen gebruiken voor al de aanwezige Models.
Dit omvat zowel Models van de applicatie zelf, als de Models uit libraries zoals oe_utils en oe_geoutils.

Hier kan eveneens de nodige generieke metadata aan ons Base Model worden meegegeven.
Hiervoor kan sqlalchemy.schema.MetaData worden gebruikt
"""
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()



# -*- coding: utf-8 -*-
import logging
import smtplib
import sys
from email import encoders
from email.header import Header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

log = logging.getLogger(__name__)
PY2 = sys.version_info[0] == 2


if PY2:
    binary_type = str
else:
    binary_type = bytes


def text_(s, encoding="latin-1", errors="strict"):
    if isinstance(s, binary_type):
        return s.decode(encoding, errors)
    return s


class SMTPClient(object):
    def __init__(self, smtp_host, sender):
        """
        Create an smtp client.

        :param smtp_host: smtp server
        :param sender: Sender of the mail
        """
        self.smtp_host = smtp_host
        self.sender = sender
        self.smtp_obj = smtplib.SMTP(self.smtp_host)

    def send_mail(
        self,
        receivers,
        subject,
        message_plain,
        message_html=None,
        cc=None,
        bcc=None,
        files=None,
    ):
        """
        Send an email.

        :param receivers: Receivers (list)
        :param subject: Email subject
        :param message_plain: Email message (plain text)
        :param message_html: Email message (html version)
        :param cc: carbon copy (list)
        :param bcc: blind carbon copy (list)
        :param files: array for attachments with
        {'name': filename, 'data': binary attachmentdata}
        """
        msg = MIMEMultipart("alternative")
        msg["Subject"] = Header(subject, "utf-8")
        msg["From"] = Header(text_(self.sender), "utf-8")
        msg["To"] = Header(text_("; ".join(receivers)), "utf-8")
        if cc:
            msg["CC"] = Header(text_("; ".join(cc)), "utf-8")
        else:
            cc = []
        if bcc:
            msg["BCC"] = Header(text_("; ".join(bcc)), "utf-8")
        else:
            bcc = []
        plain_text = MIMEText(message_plain, "plain", _charset="UTF-8")
        msg.attach(plain_text)
        if message_html:
            # According to RFC 2046, the last part of a multipart message, in this case
            # the HTML message, is best and preferred.
            html_text = MIMEText(message_html, "html", _charset="UTF-8")
            msg.attach(html_text)
        if files:
            for f in files or []:
                maintype, subtype = f["mime"].split("/", 1)
                attachment = MIMEBase(maintype, subtype)
                attachment.set_payload(f["data"])
                if not isinstance(attachment, (bytes, bytearray)):
                    log.warning("The attachment should be bytes or bytearray.")
                encoders.encode_base64(attachment)
                attachment.add_header(
                    "Content-Disposition", "attachment", filename=f["name"]
                )
                msg.attach(attachment)
        self.smtp_obj.sendmail(self.sender, receivers + cc + bcc, msg.as_string())
